BEGIN;
SELECT _v.register_patch('exchange-0009', NULL, NULL);
SET search_path TO exchange;
CREATE INDEX exchange_tables_by_pending
  ON exchange_tables (table_serial_id)
  WHERE NOT finished;
COMMENT ON INDEX exchange_tables_by_pending
  IS 'Used by exchange_do_create_tables';
CREATE FUNCTION alter_table_age_withdraw9(
 IN partition_suffix TEXT DEFAULT NULL)
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  DROP table age_withdraw;
END;
$$;
COMMENT ON FUNCTION alter_table_age_withdraw9
  IS 'Drop the table age_withdraw';
INSERT INTO exchange.exchange_tables
 (name
 ,version
 ,action
 ,partitioned
 ,by_range)
VALUES
 ('age_withdraw9'
 ,'exchange-0009'
 ,'alter'
 ,TRUE
 ,FALSE);
CREATE FUNCTION alter_table_batch_deposits9(
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  PERFORM create_partitioned_table(
    'ALTER TABLE %I'
    ' ADD COLUMN total_amount taler_amount'
    '   DEFAULT NULL'
    ';',
    'batch_deposits'
    ,''
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'total amount, NULL allowed for legacy entries (pre-1.0)'
    ,'total_amount'
    ,'batch_deposits'
    ,partition_suffix
  );
END $$;
INSERT INTO exchange_tables
    (name
    ,version
    ,action
    ,partitioned
    ,by_range)
  VALUES
    ('batch_deposits9'
    ,'exchange-0009'
    ,'alter'
    ,TRUE
    ,FALSE);
CREATE FUNCTION create_table_withdraw(
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'withdraw';
BEGIN
  PERFORM create_partitioned_table(
    'CREATE TABLE %I'
      '(withdraw_id BIGINT GENERATED BY DEFAULT AS IDENTITY'
      ',planchets_h BYTEA CONSTRAINT planchets_h_length CHECK(LENGTH(planchets_h)=64)'
      ',execution_date INT8 NOT NULL'
      ',amount_with_fee taler_amount NOT NULL'
      ',reserve_pub BYTEA NOT NULL CONSTRAINT reserve_pub_length CHECK(LENGTH(reserve_pub)=32)'
      ',reserve_sig BYTEA NOT NULL CONSTRAINT reserve_sig_length CHECK(LENGTH(reserve_sig)=64)'
      ',max_age SMALLINT CONSTRAINT max_age_positive CHECK(max_age>=0)'
      ',noreveal_index SMALLINT CONSTRAINT noreveal_index_positive CHECK(noreveal_index>=0)'
      ',selected_h BYTEA CONSTRAINT selected_h_length CHECK(LENGTH(selected_h)=64)'
      ',blinding_seed BYTEA CONSTRAINT blinding_seed_length CHECK(LENGTH(blinding_seed)>=32)'
      ',cs_r_values BYTEA[]'
      ',cs_r_choices INT8'
      ',denom_serials INT8[] NOT NULL CONSTRAINT denom_serials_array_length CHECK(cardinality(denom_serials)=cardinality(denom_sigs))'
      ',denom_sigs BYTEA[] NOT NULL CONSTRAINT denom_sigs_array_length CHECK(cardinality(denom_sigs)=cardinality(denom_serials))'
    ') %s ;'
    ,table_name
    ,'PARTITION BY HASH (reserve_pub)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
     'Commitments made when withdrawing coins and, in case of required proof of age restriction, the gamma value chosen by the exchange. '
     'It also contains the blindly signed coins, their signatures and denominations.'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'If the client explicitly commits to age-restricted coins, the gamma value chosen by the exchange in the cut-and-choose protocol; might be NULL.'
    ,'noreveal_index'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The running hash over all committed blinded planchets. Needed for recoup and (when a proof of age-restriction was required)'
     ' in the subsequent cut-and-choose protocol.'
    ,'planchets_h'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The date of execution of this withdrawal, according to the exchange'
    ,'execution_date'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'If the clients commits to age-restricted coins, the maximum age (in years) that the client explicitly commits to with this request; might be NULL.'
    ,'max_age'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Reference to the public key of the reserve from which the coins are going to be withdrawn'
    ,'reserve_pub'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Signature of the reserve''s private key over the withdraw request'
    ,'reserve_sig'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Array of references to the denominations'
    ,'denom_serials'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'In case of age restriction, the hash of the chosen (noreveal_index) blinded envelopes.'
    ,'selected_h'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Array of signatures over each blinded envelope.  If age-proof was not required, the signed envelopes are the ones'
     ' hashed into planchet_h.  Otherwise (when age-proof is required) the selected planchets (noreveal_index) were signed,'
     ' hashed into selected_h.'
    ,'denom_sigs'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The master seed for the blinding nonces, needed for blind CS signatures; maybe NULL'
    ,'blinding_seed'
    ,table_name
    ,partition_suffix
  );
 PERFORM comment_partitioned_column(
     'The pairs of R-values (calculated by the exchange) for the coins of cipher type Clause-Schnorr, based on the blinding_seed; maybe NULL.'
    ,'cs_r_values'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The bitvector of choices made by the exchange for each of the pairs in cs_r_values; maybe NULL.'
     'The vector is stored in network byte order and the lowest bit corresponds to the 0-th entry in cs_r_values (pair)'
    ,'cs_r_choices'
    ,table_name
    ,partition_suffix
  );
END
$$;
CREATE FUNCTION constrain_table_withdraw(
  IN partition_suffix TEXT
)
RETURNS void
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'withdraw';
BEGIN
  table_name = concat_ws('_', table_name, partition_suffix);
  EXECUTE FORMAT (
    'CREATE INDEX ' || table_name || '_by_reserve_pub_index '
    'ON ' || table_name || ' '
    '(reserve_pub);'
  );
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD PRIMARY KEY (reserve_pub, planchets_h);'
  );
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_withdraw_id_key'
    ' UNIQUE (withdraw_id);'
  );
END
$$;
CREATE FUNCTION foreign_table_withdraw()
RETURNS void
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'withdraw';
BEGIN
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_foreign_reserve_pub'
    ' FOREIGN KEY (reserve_pub)'
    ' REFERENCES reserves(reserve_pub) ON DELETE CASCADE;'
  );
END
$$;
CREATE FUNCTION withdraw_insert_trigger()
  RETURNS trigger
  LANGUAGE plpgsql
  AS $$
BEGIN
  INSERT INTO reserve_history
    (reserve_pub
    ,table_name
    ,serial_id)
  VALUES
    (NEW.reserve_pub
    ,'withdraw'
    ,NEW.withdraw_id);
  RETURN NEW;
END $$;
COMMENT ON FUNCTION withdraw_insert_trigger()
  IS 'Keep track of a particular withdraw in the reserve_history table.';
CREATE FUNCTION withdraw_delete_trigger()
  RETURNS trigger
  LANGUAGE plpgsql
  AS $$
BEGIN
  DELETE FROM unique_withdraw_blinding_seed
  WHERE blinding_seed = OLD.blinding_seed;
  RETURN OLD;
END $$;
COMMENT ON FUNCTION withdraw_delete_trigger()
  IS 'Delete blinding_seed from unique_withdraw_blinding_seed table.';
CREATE FUNCTION master_table_withdraw()
  RETURNS void
  LANGUAGE plpgsql
  AS $$
BEGIN
  CREATE TRIGGER withdraw_on_insert
  AFTER INSERT
  ON withdraw
  FOR EACH ROW EXECUTE FUNCTION withdraw_insert_trigger();
  CREATE TRIGGER withdraw_on_delete
  AFTER DELETE
  ON withdraw
  FOR EACH ROW EXECUTE FUNCTION withdraw_delete_trigger();
END $$;
COMMENT ON FUNCTION master_table_withdraw()
  IS 'Setup triggers to replicate withdraw into reserve_history and delete blinding_seed from unique_withdraw_blinding_seed.';
INSERT INTO exchange_tables
  (name
  ,version
  ,action
  ,partitioned
  ,by_range)
VALUES
  ('withdraw', 'exchange-0009', 'create', TRUE ,FALSE),
  ('withdraw', 'exchange-0009', 'constrain',TRUE ,FALSE),
  ('withdraw', 'exchange-0009', 'foreign', TRUE ,FALSE),
  ('withdraw', 'exchange-0009', 'master', TRUE ,FALSE);
CREATE FUNCTION create_table_unique_withdraw_blinding_seed(
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'unique_withdraw_blinding_seed';
BEGIN
  PERFORM create_partitioned_table(
    'CREATE TABLE %I'
      '(unique_withdraw_blinding_seed_id BIGINT GENERATED BY DEFAULT AS IDENTITY'
      ',blinding_seed BYTEA PRIMARY KEY'
    ') %s ;'
    ,table_name
    ,'PARTITION BY HASH (blinding_seed)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
     'Table to ensure uniqueness of the blinding_seed for CS signatures across all withdraw operations. '
    ,table_name
    ,partition_suffix
  );
END
$$;
CREATE FUNCTION constrain_table_unique_withdraw_blinding_seed(
  IN partition_suffix TEXT
)
RETURNS void
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'unique_withdraw_blinding_seed';
BEGIN
  table_name = concat_ws('_', table_name, partition_suffix);
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_withdraw_id_key'
    ' UNIQUE (unique_withdraw_blinding_seed_id);'
  );
END
$$;
INSERT INTO exchange_tables
  (name
  ,version
  ,action
  ,partitioned
  ,by_range)
VALUES
  ('unique_withdraw_blinding_seed', 'exchange-0009', 'create', TRUE ,FALSE),
  ('unique_withdraw_blinding_seed', 'exchange-0009', 'constrain',TRUE ,FALSE);
CREATE FUNCTION create_table_refresh(
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'refresh';
BEGIN
  PERFORM create_partitioned_table(
    'CREATE TABLE %I'
      '(refresh_id BIGINT GENERATED BY DEFAULT AS IDENTITY'
      ',rc BYTEA PRIMARY KEY CONSTRAINT rc_length CHECK(LENGTH(rc)=64)'
      ',execution_date INT8 NOT NULL'
      ',amount_with_fee taler_amount NOT NULL'
      ',old_coin_pub BYTEA NOT NULL'
      ',old_coin_sig BYTEA NOT NULL CHECK(LENGTH(old_coin_sig)=64)'
      ',refresh_seed BYTEA NOT NULL'
      ',noreveal_index INT4 NOT NULL  CONSTRAINT noreveal_index_positive CHECK(noreveal_index>=0)'
      ',planchets_h BYTEA CONSTRAINT planchets_h_length CHECK(LENGTH(planchets_h)=64)'
      ',selected_h BYTEA CONSTRAINT selected_h_length CHECK(LENGTH(selected_h)=64)'
      ',blinding_seed BYTEA CONSTRAINT blinding_seed_length CHECK(LENGTH(blinding_seed)>=32)'
      ',cs_r_values BYTEA[]'
      ',cs_r_choices INT8'
      ',denom_serials INT8[] NOT NULL CONSTRAINT denom_serials_array_length CHECK(cardinality(denom_serials)=cardinality(denom_sigs))'
      ',denom_sigs BYTEA[] NOT NULL CONSTRAINT denom_sigs_array_length CHECK(cardinality(denom_sigs)=cardinality(denom_serials))'
    ') %s ;'
    ,table_name
    ,'PARTITION BY HASH (rc)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
    'The data provided by the client for the melting operation of an old coin and he choices made by the exchange '
    ' with respect to the cut-and-choose protocol: nonreveal_index and the corresponding chosen'
    ' blinded coin envelope along with the denomination signatures at the moment of the melting.'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The hash over the refresh request, which serves as the primary key'
     ' for the lookup during the reveal phase.'
    ,'rc'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The publice nonce from which all other nonces for all n*kappa coin candidates are derived for which'
     ' the old coin proves ownership via signatures'
    ,'refresh_seed'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The gamma value chosen by the exchange in the cut-and-choose protocol'
    ,'noreveal_index'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The date of execution of the melting operation, according to the exchange'
    ,'execution_date'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Reference to the public key of the old coin which is melted'
    ,'old_coin_pub'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Signature of the old coin''s private key over the melt request'
    ,'old_coin_sig'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Array of references to the denominations'
    ,'denom_serials'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The master seed for the blinding nonces, needed for blind CS signatures; maybe NULL'
    ,'blinding_seed'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The public pairs of R-values provided by the exchange for the CS denominations; might be NULL'
    ,'cs_r_values'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The bitvector of choices made by the exchange for each of the pairs in cs_r_values; maybe NULL.'
     'The vector is stored in network byte order and the lowest bit corresponds to the 0-th entry in cs_r_values (pair)'
    ,'cs_r_choices'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The hash over all kappa*n blinded planchets that were provided by the client'
    ,'planchets_h'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'The hash over the n blinded planchets that were selected by the exchange.'
    ,'selected_h'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Array of signatures, one for each blinded envelope'
    ,'denom_sigs'
    ,table_name
    ,partition_suffix
  );
END
$$;
CREATE FUNCTION constrain_table_refresh(
  IN partition_suffix TEXT
)
RETURNS void
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'refresh';
BEGIN
  table_name = concat_ws('_', table_name, partition_suffix);
  EXECUTE FORMAT (
    'CREATE INDEX ' || table_name || '_by_old_coin_pub_index '
    'ON ' || table_name || ' '
    '(old_coin_pub);'
  );
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_refresh_id_key'
    ' UNIQUE (refresh_id);'
  );
END
$$;
CREATE FUNCTION foreign_table_refresh()
RETURNS void
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'refresh';
BEGIN
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_foreign_coin_pub'
    ' FOREIGN KEY (old_coin_pub) '
    ' REFERENCES known_coins (coin_pub) ON DELETE CASCADE'
  );
END
$$;
CREATE FUNCTION refresh_insert_trigger()
  RETURNS trigger
  LANGUAGE plpgsql
  AS $$
BEGIN
  INSERT INTO coin_history
    (coin_pub
    ,table_name
    ,serial_id)
  VALUES
    (NEW.old_coin_pub
    ,'refresh'
    ,NEW.refresh_id);
  RETURN NEW;
END $$;
COMMENT ON FUNCTION refresh_insert_trigger()
  IS 'Keep track of a particular refresh in the coin_history table.';
CREATE FUNCTION refresh_delete_trigger()
  RETURNS trigger
  LANGUAGE plpgsql
  AS $$
BEGIN
  DELETE FROM unique_refresh_blinding_seed
  WHERE blinding_seed = OLD.blinding_seed;
  RETURN OLD;
END $$;
COMMENT ON FUNCTION refresh_delete_trigger()
  IS 'Delete blinding_seed from unique_refresh_blinding_seed table.';
CREATE FUNCTION master_table_refresh()
  RETURNS void
  LANGUAGE plpgsql
  AS $$
BEGIN
  CREATE TRIGGER refresh_on_insert
  AFTER INSERT
  ON refresh
  FOR EACH ROW EXECUTE FUNCTION refresh_insert_trigger();
  CREATE TRIGGER refresh_on_delete
  AFTER DELETE
  ON refresh
  FOR EACH ROW EXECUTE FUNCTION refresh_delete_trigger();
END $$;
COMMENT ON FUNCTION master_table_refresh()
  IS 'Setup triggers to replicate refresh into coin_history and delete blinding_seed from unique_refresh_blinding_seed.';
INSERT INTO exchange_tables
  (name
  ,version
  ,action
  ,partitioned
  ,by_range)
VALUES
  ('refresh', 'exchange-0009', 'create', TRUE ,FALSE),
  ('refresh', 'exchange-0009', 'constrain',TRUE ,FALSE),
  ('refresh', 'exchange-0009', 'foreign', TRUE ,FALSE),
  ('refresh', 'exchange-0009', 'master', TRUE ,FALSE);
CREATE FUNCTION create_table_unique_refresh_blinding_seed(
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'unique_refresh_blinding_seed';
BEGIN
  PERFORM create_partitioned_table(
    'CREATE TABLE %I'
      '(unique_refresh_blinding_seed_id BIGINT GENERATED BY DEFAULT AS IDENTITY'
      ',blinding_seed BYTEA PRIMARY KEY'
    ') %s ;'
    ,table_name
    ,'PARTITION BY HASH (blinding_seed)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
     'Table to ensure uniqueness of the blinding_seed for CS signatures across all refresh operations. '
    ,table_name
    ,partition_suffix
  );
END
$$;
CREATE FUNCTION constrain_table_unique_refresh_blinding_seed(
  IN partition_suffix TEXT
)
RETURNS void
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'unique_refresh_blinding_seed';
BEGIN
  table_name = concat_ws('_', table_name, partition_suffix);
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_refresh_id_key'
    ' UNIQUE (unique_refresh_blinding_seed_id);'
  );
END
$$;
INSERT INTO exchange_tables
  (name
  ,version
  ,action
  ,partitioned
  ,by_range)
VALUES
  ('unique_refresh_blinding_seed', 'exchange-0009', 'create', TRUE ,FALSE),
  ('unique_refresh_blinding_seed', 'exchange-0009', 'constrain',TRUE ,FALSE);
CREATE FUNCTION alter_table_recoup2()
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  RAISE NOTICE 'Dropping old recoup tables';
  DROP TABLE recoup_by_reserve;
  DROP TABLE recoup;
  ALTER FUNCTION create_table_recoup9(TEXT)
   RENAME TO create_table_recoup;
  ALTER FUNCTION create_table_recoup_by_reserve9(TEXT)
   RENAME TO create_table_recoup_by_reserve;
END
$$;
CREATE FUNCTION create_table_recoup9(
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'recoup';
BEGIN
  PERFORM create_partitioned_table(
    'CREATE TABLE %I'
      '(recoup_uuid BIGINT GENERATED BY DEFAULT AS IDENTITY'
      ',coin_pub BYTEA NOT NULL CHECK (LENGTH(coin_pub)=32)'
      ',coin_sig BYTEA NOT NULL CHECK(LENGTH(coin_sig)=64)'
      ',coin_blind BYTEA NOT NULL CHECK(LENGTH(coin_blind)=32)'
      ',amount taler_amount NOT NULL'
      ',recoup_timestamp INT8 NOT NULL'
      ',withdraw_id INT8 NOT NULL'
    ') %s ;'
    ,table_name
    ,'PARTITION BY HASH (coin_pub);'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
     'Information about recoups that were executed between a coin and a reserve. In this type of recoup, the amount is credited back to the reserve from which the coin originated.'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Coin that is being debited in the recoup. Do not CASCADE ON DROP on the coin_pub, as we may keep the coin alive!'
    ,'coin_pub'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Identifies the h_commitment of the recouped coin and provides the link to the credited reserve.'
    ,'withdraw_id'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Signature by the coin affirming the recoup, of type TALER_SIGNATURE_WALLET_COIN_RECOUP'
    ,'coin_sig'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Denomination blinding key used when creating the blinded coin from the planchet. Secret revealed during the recoup to provide the linkage between the coin and the withdraw operation.'
    ,'coin_blind'
    ,table_name
    ,partition_suffix
  );
END
$$;
CREATE FUNCTION constrain_table_recoup9(
  IN partition_suffix TEXT
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'recoup';
BEGIN
  table_name = concat_ws('_', table_name, partition_suffix);
  EXECUTE FORMAT (
    'CREATE INDEX ' || table_name || '_by_coin_pub_index '
    'ON ' || table_name || ' '
    '(coin_pub);'
  );
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_recoup_uuid_key'
    ' UNIQUE (recoup_uuid) '
  );
END
$$;
CREATE OR REPLACE FUNCTION foreign_table_recoup9()
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'recoup';
BEGIN
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_foreign_withdraw'
    ' FOREIGN KEY (withdraw_id) '
    ' REFERENCES withdraw (withdraw_id) ON DELETE CASCADE'
    ',ADD CONSTRAINT ' || table_name || '_foreign_coin_pub'
    ' FOREIGN KEY (coin_pub) '
    ' REFERENCES known_coins (coin_pub)'
  );
END
$$;
CREATE FUNCTION create_table_recoup_by_reserve9(
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'recoup_by_reserve';
BEGIN
  PERFORM create_partitioned_table(
  'CREATE TABLE %I'
    '(withdraw_id INT8 NOT NULL' 
    ',coin_pub BYTEA CHECK (LENGTH(coin_pub)=32)' 
    ') %s ;'
    ,table_name
    ,'PARTITION BY HASH (withdraw_id)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
     'Information in this table is strictly redundant with that of recoup, but saved by a different primary key for fast lookups by withdraw_id.'
    ,table_name
    ,partition_suffix
  );
END
$$;
CREATE FUNCTION constrain_table_recoup_by_reserve9(
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'recoup_by_reserve';
BEGIN
  table_name = concat_ws('_', table_name, partition_suffix);
  EXECUTE FORMAT (
    'CREATE INDEX ' || table_name || '_main_index '
    'ON ' || table_name || ' '
    '(withdraw_id);'
  );
END
$$;
CREATE OR REPLACE FUNCTION recoup_insert_trigger()
  RETURNS trigger
  LANGUAGE plpgsql
  AS $$
BEGIN
  INSERT INTO recoup_by_reserve
    (withdraw_id
    ,coin_pub)
  VALUES
    (NEW.withdraw_id
    ,NEW.coin_pub);
  INSERT INTO coin_history
    (coin_pub
    ,table_name
    ,serial_id)
  VALUES
    (NEW.coin_pub
    ,'recoup'
    ,NEW.recoup_uuid);
  INSERT INTO reserve_history
    (reserve_pub
    ,table_name
    ,serial_id)
  SELECT
     res.reserve_pub
    ,'recoup'
    ,NEW.recoup_uuid
  FROM withdraw wd
  JOIN reserves res
    USING (reserve_pub)
    WHERE wd.withdraw_id = NEW.withdraw_id;
  RETURN NEW;
END $$;
COMMENT ON FUNCTION recoup_insert_trigger()
  IS 'Replicates recoup inserts into recoup_by_reserve table and updates the coin_history table.';
CREATE OR REPLACE FUNCTION recoup_delete_trigger()
  RETURNS trigger
  LANGUAGE plpgsql
  AS $$
BEGIN
  DELETE FROM recoup_by_reserve
   WHERE withdraw_id = OLD.withdraw_id
     AND coin_pub = OLD.coin_pub;
  RETURN OLD;
END $$;
COMMENT ON FUNCTION recoup_delete_trigger()
  IS 'Replicate recoup deletions into recoup_by_reserve table.';
CREATE FUNCTION master_table_recoup9()
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  CREATE TRIGGER recoup_on_insert
    AFTER INSERT
     ON recoup
     FOR EACH ROW EXECUTE FUNCTION recoup_insert_trigger();
  CREATE TRIGGER recoup_on_delete
    AFTER DELETE
      ON recoup
     FOR EACH ROW EXECUTE FUNCTION recoup_delete_trigger();
END
$$;
INSERT INTO exchange_tables
    (name
    ,version
    ,action
    ,partitioned
    ,by_range)
  VALUES
    ('recoup2'
    ,'exchange-0009'
    ,'alter'
    ,TRUE
    ,FALSE),
    ('recoup' 
    ,'exchange-0009'
    ,'create'
    ,TRUE
    ,FALSE),
    ('recoup9'
    ,'exchange-0009'
    ,'constrain'
    ,TRUE
    ,FALSE),
    ('recoup9'
    ,'exchange-0009'
    ,'foreign'
    ,TRUE
    ,FALSE),
    ('recoup_by_reserve' 
    ,'exchange-0009'
    ,'create'
    ,TRUE
    ,FALSE),
    ('recoup_by_reserve9'
    ,'exchange-0009'
    ,'constrain'
    ,TRUE
    ,FALSE),
    ('recoup9'
    ,'exchange-0009'
    ,'master'
    ,TRUE
    ,FALSE);
CREATE FUNCTION alter_table_recoup_refresh2()
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  RAISE NOTICE 'Dropping old recoup_refresh table';
  DROP TABLE recoup_refresh;
  ALTER FUNCTION create_table_recoup_refresh9(TEXT)
   RENAME TO create_table_recoup_refresh;
END
$$;
CREATE FUNCTION create_table_recoup_refresh9(
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'recoup_refresh';
BEGIN
  PERFORM create_partitioned_table(
    'CREATE TABLE %I'
    '(recoup_refresh_uuid BIGINT GENERATED BY DEFAULT AS IDENTITY'
    ',coin_pub BYTEA NOT NULL CHECK (LENGTH(coin_pub)=32)'
    ',known_coin_id BIGINT NOT NULL'
    ',coin_sig BYTEA NOT NULL CHECK(LENGTH(coin_sig)=64)'
    ',coin_blind BYTEA NOT NULL CHECK(LENGTH(coin_blind)=32)'
    ',amount taler_amount NOT NULL'
    ',recoup_timestamp INT8 NOT NULL'
    ',refresh_id INT8 NOT NULL'
    ') %s ;'
    ,table_name
    ,'PARTITION BY HASH (coin_pub)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
     'Table of coins that originated from a refresh operation and that were recouped. Links the (fresh) coin to the melted operation (and thus the old coin). A recoup on a refreshed coin credits the old coin and debits the fresh coin.'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Refreshed coin of a revoked denomination where the residual value is credited to the old coin. Do not CASCADE ON DROP on the coin_pub, as we may keep the coin alive!'
    ,'coin_pub'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Used for garbage collection (in the absence of foreign constraints, in the future)'
    ,'known_coin_id'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Link to the refresh operation. Also identifies the h_blind_ev of the recouped coin (as h_coin_ev).'
    ,'refresh_id'
    ,table_name
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'Denomination blinding key used when creating the blinded coin from the planchet. Secret revealed during the recoup to provide the linkage between the coin and the refresh operation.'
    ,'coin_blind'
    ,table_name
    ,partition_suffix
  );
END
$$;
CREATE FUNCTION constrain_table_recoup_refresh9(
  IN partition_suffix TEXT
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'recoup_refresh';
BEGIN
  table_name = concat_ws('_', table_name, partition_suffix);
  EXECUTE FORMAT (
    'CREATE INDEX ' || table_name || '_by_refresh_id_index'
    ' ON ' || table_name || ' '
    '(refresh_id);'
  );
  EXECUTE FORMAT (
    'COMMENT ON INDEX ' || table_name || '_by_refresh_id_index '
    'IS ' || quote_literal('used in exchange_do_melt for zombie coins (rare)') || ';'
  );
  EXECUTE FORMAT (
    'CREATE INDEX ' || table_name || '_by_coin_pub_index'
    ' ON ' || table_name || ' '
    '(coin_pub);'
  );
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_recoup_refresh_uuid_key'
    ' UNIQUE (recoup_refresh_uuid) '
  );
END
$$;
CREATE FUNCTION foreign_table_recoup_refresh9()
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'recoup_refresh';
BEGIN
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_foreign_coin_pub'
    ' FOREIGN KEY (coin_pub) '
    ' REFERENCES known_coins (coin_pub)'
    ',ADD CONSTRAINT ' || table_name || '_foreign_known_coin_id'
    ' FOREIGN KEY (known_coin_id) '
    ' REFERENCES known_coins (known_coin_id) ON DELETE CASCADE'
    ',ADD CONSTRAINT ' || table_name || '_foreign_refresh_id'
    ' FOREIGN KEY (refresh_id) '
    ' REFERENCES refresh (refresh_id) ON DELETE CASCADE'
  );
END
$$;
CREATE OR REPLACE FUNCTION recoup_refresh_insert_trigger()
  RETURNS trigger
  LANGUAGE plpgsql
  AS $$
BEGIN
  INSERT INTO exchange.coin_history
    (coin_pub
    ,table_name
    ,serial_id)
  VALUES
     (NEW.coin_pub
    ,'recoup_refresh::NEW'
    ,NEW.recoup_refresh_uuid);
  INSERT INTO exchange.coin_history
    (coin_pub
    ,table_name
    ,serial_id)
  SELECT
     refresh.old_coin_pub
    ,'recoup_refresh::OLD'
    ,NEW.recoup_refresh_uuid
    FROM refresh
    WHERE refresh.refresh_id = NEW.refresh_id;
  RETURN NEW;
END $$;
COMMENT ON FUNCTION coin_deposits_insert_trigger()
  IS 'Automatically generate coin history entry.';
CREATE FUNCTION master_table_recoup_refresh9()
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  CREATE TRIGGER recoup_refresh_on_insert
    AFTER INSERT
     ON recoup_refresh
     FOR EACH ROW EXECUTE FUNCTION recoup_refresh_insert_trigger();
END $$;
INSERT INTO exchange_tables
    (name
    ,version
    ,action
    ,partitioned
    ,by_range)
  VALUES
    ('recoup_refresh2'
    ,'exchange-0009'
    ,'alter'
    ,TRUE
    ,FALSE),
    ('recoup_refresh' 
    ,'exchange-0009'
    ,'create'
    ,TRUE
    ,FALSE),
    ('recoup_refresh9'
    ,'exchange-0009'
    ,'constrain'
    ,TRUE
    ,FALSE),
    ('recoup_refresh9'
    ,'exchange-0009'
    ,'foreign'
    ,TRUE
    ,FALSE),
    ('recoup_refresh9'
    ,'exchange-0009'
    ,'master'
    ,TRUE
    ,FALSE);
CREATE TYPE statistic_range AS
  ENUM('century', 'decade', 'year', 'quarter', 'month', 'week', 'day', 'hour', 'minute', 'second');
CREATE TYPE statistic_type AS
  ENUM('amount', 'number');
CREATE TABLE exchange_statistic_bucket_meta
  (bmeta_serial_id INT8 GENERATED BY DEFAULT AS IDENTITY PRIMARY KEY
  ,origin TEXT NOT NULL
  ,slug TEXT NOT NULL
  ,description TEXT NOT NULL
  ,stype statistic_type NOT NULL
  ,ranges statistic_range[] NOT NULL
  ,ages INT4[] NOT NULL
  ,UNIQUE(slug,stype)
  ,CONSTRAINT equal_array_length
    CHECK (array_length(ranges,1) =
           array_length(ages,1))
  );
COMMENT ON TABLE exchange_statistic_bucket_meta
  IS 'meta data about a statistic with events falling into buckets we are tracking';
COMMENT ON COLUMN exchange_statistic_bucket_meta.bmeta_serial_id
  IS 'unique identifier for this type of bucket statistic we are tracking';
COMMENT ON COLUMN exchange_statistic_bucket_meta.origin
  IS 'which customization schema does this statistic originate from (used for easy deletion)';
COMMENT ON COLUMN exchange_statistic_bucket_meta.slug
  IS 'keyword (or name) of the statistic; identifies what the statistic is about; should be a slug suitable for a URI path';
COMMENT ON COLUMN exchange_statistic_bucket_meta.description
  IS 'description of the statistic being tracked';
COMMENT ON COLUMN exchange_statistic_bucket_meta.stype
  IS 'statistic type, what kind of data is being tracked, amount or number';
COMMENT ON COLUMN exchange_statistic_bucket_meta.ranges
  IS 'size of the buckets that are being kept for this statistic';
COMMENT ON COLUMN exchange_statistic_bucket_meta.ages
  IS 'determines how long into the past we keep buckets for the range at the given index around (in generations)';
CREATE INDEX exchange_statistic_bucket_meta_by_origin
  ON exchange_statistic_bucket_meta
  (origin);
CREATE FUNCTION create_table_exchange_statistic_bucket_counter (
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  PERFORM create_partitioned_table (
    'CREATE TABLE %I'
      '(bmeta_serial_id INT8 NOT NULL'
      ' REFERENCES exchange_statistic_bucket_meta (bmeta_serial_id) ON DELETE CASCADE'
      ',h_payto BYTEA CHECK (LENGTH(h_payto)=32)'
      ',bucket_start INT8 NOT NULL'
      ',bucket_range statistic_range NOT NULL'
      ',cumulative_number INT8 NOT NULL'
      ',UNIQUE (h_payto,bmeta_serial_id,bucket_start,bucket_range)'
    ') %s;'
    ,'exchange_statistic_bucket_counter'
    ,'PARTITION BY HASH (h_payto)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
     'various numeric statistics (cumulative counters) being tracked by bucket into which they fall'
    ,'exchange_statistic_bucket_counter'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'identifies what the statistic is about'
    ,'bmeta_serial_id'
    ,'exchange_statistic_bucket_counter'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'identifies an account (hash of normalized payto) for which the statistic is kept, NULL for global statistics'
    ,'h_payto'
    ,'exchange_statistic_bucket_counter'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'start date for the bucket in seconds since the epoch'
    ,'bucket_start'
    ,'exchange_statistic_bucket_counter'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'range of the bucket'
    ,'bucket_range'
    ,'exchange_statistic_bucket_counter'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'aggregate (sum) of tracked by the statistic; what exactly is tracked is determined by the keyword'
    ,'cumulative_number'
    ,'exchange_statistic_bucket_counter'
    ,partition_suffix
  );
END $$;
CREATE FUNCTION create_table_exchange_statistic_bucket_amount (
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  PERFORM create_partitioned_table (
    'CREATE TABLE %I'
      '(bmeta_serial_id INT8 NOT NULL'
       ' REFERENCES exchange_statistic_bucket_meta (bmeta_serial_id) ON DELETE CASCADE'
      ',h_payto BYTEA CHECK (LENGTH(h_payto)=32)'
      ',bucket_start INT8 NOT NULL'
      ',bucket_range statistic_range NOT NULL'
      ',cumulative_value taler_amount NOT NULL'
      ',CHECK ((cumulative_value).val IS NOT NULL AND (cumulative_value).frac IS NOT NULL)'
      ',UNIQUE (h_payto,bmeta_serial_id,bucket_start,bucket_range)'
    ') %s;'
    ,'exchange_statistic_bucket_amount'
    ,'PARTITION BY HASH(h_payto)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table (
     'various amount statistics being tracked'
    ,'exchange_statistic_bucket_amount'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'identifies what the statistic is about'
    ,'bmeta_serial_id'
    ,'exchange_statistic_bucket_amount'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column (
     'identifies an account (hash of normalized payto) for which the statistic is kept, NULL for global statistics'
    ,'h_payto'
    ,'exchange_statistic_bucket_amount'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'start date for the bucket in seconds since the epoch'
    ,'bucket_start'
    ,'exchange_statistic_bucket_amount'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'range of the bucket'
    ,'bucket_range'
    ,'exchange_statistic_bucket_amount'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
     'amount being tracked'
    ,'cumulative_value'
    ,'exchange_statistic_bucket_amount'
    ,partition_suffix
  );
END $$;
CREATE TABLE exchange_statistic_interval_meta
  (imeta_serial_id INT8 GENERATED BY DEFAULT AS IDENTITY PRIMARY KEY
  ,origin TEXT NOT NULL
  ,slug TEXT NOT NULL
  ,description TEXT NOT NULL
  ,stype statistic_type NOT NULL
  ,ranges INT8[] NOT NULL CHECK (array_length(ranges,1) > 0)
  ,precisions INT8[] NOT NULL CHECK (array_length(precisions,1) > 0)
  ,UNIQUE(slug,stype)
  ,CONSTRAINT equal_array_length
    CHECK (array_length(ranges,1) =
           array_length(precisions,1))
  );
COMMENT ON TABLE exchange_statistic_interval_meta
  IS 'meta data about an interval statistic we are tracking';
COMMENT ON COLUMN exchange_statistic_interval_meta.imeta_serial_id
  IS 'unique identifier for this type of interval statistic we are tracking';
COMMENT ON COLUMN exchange_statistic_interval_meta.origin
  IS 'which customization schema does this statistic originate from (used for easy deletion)';
COMMENT ON COLUMN exchange_statistic_interval_meta.slug
  IS 'keyword (or name) of the statistic; identifies what the statistic is about; should be a slug suitable for a URI path';
COMMENT ON COLUMN exchange_statistic_interval_meta.description
  IS 'description of the statistic being tracked';
COMMENT ON COLUMN exchange_statistic_interval_meta.stype
  IS 'statistic type, what kind of data is being tracked, amount or number';
COMMENT ON COLUMN exchange_statistic_interval_meta.ranges
  IS 'range of values that is being kept for this statistic, in seconds, must be monotonically increasing';
COMMENT ON COLUMN exchange_statistic_interval_meta.precisions
  IS 'determines how precisely we track which events fall into the range at the same index (allowing us to coalesce events with timestamps in proximity close to the given precision), in seconds, 0 is not allowed';
CREATE INDEX exchange_statistic_interval_meta_by_origin
  ON exchange_statistic_interval_meta
  (origin);
CREATE FUNCTION create_table_exchange_statistic_counter_event (
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  PERFORM create_partitioned_table(
    'CREATE TABLE %I'
      '(nevent_serial_id INT8 GENERATED BY DEFAULT AS IDENTITY'
      ',imeta_serial_id INT8'
      ' REFERENCES exchange_statistic_interval_meta (imeta_serial_id) ON DELETE CASCADE'
      ',h_payto BYTEA CHECK (LENGTH(h_payto)=32)'
      ',slot INT8 NOT NULL'
      ',delta INT8 NOT NULL'
      ',UNIQUE (h_payto,imeta_serial_id,slot)'
    ') %s ;'
    ,'exchange_statistic_counter_event'
    ,'PARTITION BY HASH(h_payto)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
    'number to decrement an interval statistic by when a certain time value is reached'
    ,'exchange_statistic_counter_event'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'unique identifier for this number event'
    ,'nevent_serial_id'
    ,'exchange_statistic_counter_event'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'identifies what the statistic is about; must be of stype number'
    ,'imeta_serial_id'
    ,'exchange_statistic_counter_event'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'identifies an account (hash of normalized payto) for which the statistic is kept, NULL for global statistics'
    ,'h_payto'
    ,'exchange_statistic_counter_event'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'identifies the time slot at which the given event(s) happened, rounded down by the respective precisions value'
    ,'slot'
    ,'exchange_statistic_counter_event'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'total cumulative number that was added at the time identified by slot'
    ,'delta'
    ,'exchange_statistic_counter_event'
    ,partition_suffix
  );
END $$;
CREATE FUNCTION constrain_table_exchange_statistic_counter_event(
  IN partition_suffix TEXT
)
RETURNS void
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT default 'exchange_statistic_counter_event';
BEGIN
  table_name = concat_ws('_', table_name, partition_suffix);
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_nevent_serial_id_key'
    ' UNIQUE (nevent_serial_id)'
  );
END $$;
CREATE FUNCTION create_table_exchange_statistic_interval_counter (
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  PERFORM create_partitioned_table(
    'CREATE TABLE %I'
      '(imeta_serial_id INT8 NOT NULL'
      ' REFERENCES exchange_statistic_interval_meta (imeta_serial_id) ON DELETE CASCADE'
      ',h_payto BYTEA CHECK (LENGTH(h_payto)=32)'
      ',range INT8 NOT NULL'
      ',event_delimiter INT8 NOT NULL'
      ',cumulative_number INT8 NOT NULL'
      ',UNIQUE (h_payto,imeta_serial_id,range)'
    ') %s ;'
    ,'exchange_statistic_interval_counter'
    ,'PARTITION BY HASH(h_payto)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
    'various numeric statistics (cumulative counters) being tracked'
    ,'exchange_statistic_interval_counter'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'identifies what the statistic is about'
    ,'imeta_serial_id'
    ,'exchange_statistic_interval_counter'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'identifies an account (hash of normalized payto) for which the statistic is kept, NULL for global statistics'
    ,'h_payto'
    ,'exchange_statistic_interval_counter'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'for which range is this the counter; note that the cumulative_number excludes the values already stored in smaller ranges'
    ,'range'
    ,'exchange_statistic_interval_counter'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'determines the last event currently included in the interval'
    ,'event_delimiter'
    ,'exchange_statistic_interval_counter'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'aggregate (sum) of tracked by the statistic; what exactly is tracked is determined by the keyword'
    ,'cumulative_number'
    ,'exchange_statistic_interval_counter'
    ,partition_suffix
  );
END $$;
CREATE FUNCTION foreign_table_exchange_statistic_interval_counter()
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'exchange_statistic_interval_counter';
BEGIN
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_event_delimiter_foreign_key'
    ' FOREIGN KEY (event_delimiter) '
    ' REFERENCES exchange_statistic_counter_event (nevent_serial_id) ON DELETE RESTRICT'
  );
END $$;
CREATE FUNCTION create_table_exchange_statistic_amount_event (
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  PERFORM create_partitioned_table(
    'CREATE TABLE %I'
      '(aevent_serial_id INT8 GENERATED BY DEFAULT AS IDENTITY'
      ',imeta_serial_id INT8'
      ' REFERENCES exchange_statistic_interval_meta (imeta_serial_id) ON DELETE CASCADE'
      ',h_payto BYTEA CHECK (LENGTH(h_payto)=32)'
      ',slot INT8 NOT NULL'
      ',delta taler_amount NOT NULL'
      ',CHECK ((delta).val IS NOT NULL AND (delta).frac IS NOT NULL)'
      ',CONSTRAINT event_key UNIQUE (h_payto,imeta_serial_id,slot)'
    ') %s ;'
    ,'exchange_statistic_amount_event'
    ,'PARTITION BY HASH(h_payto)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
    'amount to decrement an interval statistic by when a certain time value is reached'
    ,'exchange_statistic_amount_event'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'unique identifier for this amount event'
    ,'aevent_serial_id'
    ,'exchange_statistic_amount_event'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'identifies what the statistic is about; must be of clazz interval and of stype amount'
    ,'imeta_serial_id'
    ,'exchange_statistic_amount_event'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'identifies an account (hash of normalized payto) for which the statistic is kept, NULL for global statistics'
    ,'h_payto'
    ,'exchange_statistic_amount_event'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'identifies the time slot at which the given event(s) happened'
    ,'slot'
    ,'exchange_statistic_amount_event'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'total cumulative amount that was added at the time identified by slot'
    ,'delta'
    ,'exchange_statistic_amount_event'
    ,partition_suffix
  );
END $$;
CREATE FUNCTION constrain_table_exchange_statistic_amount_event(
  IN partition_suffix TEXT
)
RETURNS void
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT default 'exchange_statistic_amount_event';
BEGIN
  table_name = concat_ws('_', table_name, partition_suffix);
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_aevent_serial_id_key'
    ' UNIQUE (aevent_serial_id)'
  );
END $$;
CREATE FUNCTION create_table_exchange_statistic_interval_amount (
  IN partition_suffix TEXT DEFAULT NULL
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
BEGIN
  PERFORM create_partitioned_table(
    'CREATE TABLE %I'
      '(imeta_serial_id INT8 NOT NULL'
      ' REFERENCES exchange_statistic_interval_meta (imeta_serial_id) ON DELETE CASCADE'
      ',h_payto BYTEA CHECK (LENGTH(h_payto)=32)'
      ',event_delimiter INT8 NOT NULL'
      ',range INT8 NOT NULL'
      ',cumulative_value taler_amount NOT NULL'
      ',CHECK ((cumulative_value).val IS NOT NULL AND (cumulative_value).frac IS NOT NULL)'
      ',UNIQUE (h_payto,imeta_serial_id,range)'
    ') %s ;'
    ,'exchange_statistic_interval_amount'
    ,'PARTITION BY HASH(h_payto)'
    ,partition_suffix
  );
  PERFORM comment_partitioned_table(
     'various amount statistics being tracked'
    ,'exchange_statistic_interval_amount'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'identifies what the statistic is about'
    ,'imeta_serial_id'
    ,'exchange_statistic_interval_amount'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'identifies an account (hash of normalized payto) for which the statistic is kept, NULL for global statistics'
    ,'h_payto'
    ,'exchange_statistic_interval_amount'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'for which range is this the counter; note that the cumulative_number excludes the values already stored in smaller ranges'
    ,'range'
    ,'exchange_statistic_interval_amount'
    ,partition_suffix
  );
  PERFORM comment_partitioned_column(
    'amount affected by the event'
    ,'cumulative_value'
    ,'exchange_statistic_interval_amount'
    ,partition_suffix
  );
END $$;
CREATE FUNCTION foreign_table_exchange_statistic_interval_amount()
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'exchange_statistic_interval_amount';
BEGIN
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_event_delimiter_foreign_key'
    ' FOREIGN KEY (event_delimiter) '
    ' REFERENCES exchange_statistic_amount_event (aevent_serial_id) ON DELETE RESTRICT'
  );
END $$;
CREATE TYPE exchange_statistic_interval_number_get_return_value
  AS
  (range INT8
  ,rvalue INT8
  );
COMMENT ON TYPE exchange_statistic_interval_number_get_return_value
  IS 'Return type for exchange_statistic_interval_number_get stored procedure';
CREATE TYPE exchange_statistic_interval_amount_get_return_value
  AS
  (range INT8
  ,rvalue taler_amount
  );
COMMENT ON TYPE exchange_statistic_interval_amount_get_return_value
  IS 'Return type for exchange_statistic_interval_amount_get stored procedure';
INSERT INTO exchange_tables
    (name
    ,version
    ,action
    ,partitioned
    ,by_range)
  VALUES
    ('exchange_statistic_bucket_counter'
    ,'exchange-0009'
    ,'create'
    ,TRUE
    ,FALSE),
    ('exchange_statistic_bucket_amount'
    ,'exchange-0009'
    ,'create'
    ,TRUE
    ,FALSE),
    ('exchange_statistic_counter_event'
    ,'exchange-0009'
    ,'create'
    ,TRUE
    ,FALSE),
    ('exchange_statistic_counter_event'
    ,'exchange-0009'
    ,'constrain'
    ,TRUE
    ,FALSE),
    ('exchange_statistic_interval_counter'
    ,'exchange-0009'
    ,'create'
    ,TRUE
    ,FALSE),
    ('exchange_statistic_interval_counter'
    ,'exchange-0009'
    ,'foreign'
    ,TRUE
    ,FALSE),
    ('exchange_statistic_amount_event'
    ,'exchange-0009'
    ,'create'
    ,TRUE
    ,FALSE),
    ('exchange_statistic_amount_event'
    ,'exchange-0009'
    ,'constrain'
    ,TRUE
    ,FALSE),
    ('exchange_statistic_interval_amount'
    ,'exchange-0009'
    ,'create'
    ,TRUE
    ,FALSE),
    ('exchange_statistic_interval_amount'
    ,'exchange-0009'
    ,'foreign'
    ,TRUE
    ,FALSE);
CREATE FUNCTION foreign_table_aggregation_transient9()
RETURNS void
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'aggregation_transient';
BEGIN
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_foreign_wire_target_h_payto'
    ' FOREIGN KEY (wire_target_h_payto) '
    ' REFERENCES wire_targets (wire_target_h_payto) ON DELETE RESTRICT'
  );
END
$$;
CREATE FUNCTION constrain_table_aggregation_transient9(
  IN partition_suffix TEXT
)
RETURNS VOID
LANGUAGE plpgsql
AS $$
DECLARE
  table_name TEXT DEFAULT 'aggregation_transient';
BEGIN
  table_name = concat_ws('_', table_name, partition_suffix);
  EXECUTE FORMAT (
    'ALTER TABLE ' || table_name ||
    ' ADD CONSTRAINT ' || table_name || '_wire_target_h_payto_and_wtid_unique'
    '   UNIQUE (wire_target_h_payto,wtid_raw)'
  );
END $$;
INSERT INTO exchange_tables
    (name
    ,version
    ,action
    ,partitioned
    ,by_range)
  VALUES
    ('aggregation_transient9'
    ,'exchange-0009'
    ,'foreign'
    ,TRUE
    ,FALSE),
    ('aggregation_transient9'
    ,'exchange-0009'
    ,'constrain'
    ,TRUE
    ,FALSE);
COMMIT;
